package editor;
import gui.CanvasDialogue;
import gui.ImageArray;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.util.LinkedHashMap;

import javax.swing.BorderFactory;
import javax.swing.DefaultListCellRenderer;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import javax.swing.SwingUtilities;
import javax.swing.Timer;

import utils.Data;
import utils.Global;

/**
 * Builds a non-resizable confirmation dialogue with a dynamic central panel containing a list
 * of items (icons and text in this case). The list changes depending on the contents of a 
 * text field at the bottom of the dialogue. A default item list is displayed should the text
 * field be emptied. Furthermore, a previsualisation image is displayed whenever the mouse hovers
 * over the item in question. Lastly, the top section of the window is populated with buttons aligned
 * with both corners.
 * @author hqcasanova
 * 
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/.
 */
public class SelectDoc extends CanvasDialogue implements ActionListener,
														  KeyListener,
														  WindowListener,
														  MouseMotionListener,
														  MouseListener,
														  Runnable {
	//Other windows this one links to
	private final Editor editor;
	private final Login login;
	
	//Interactive elements
	private JButton newButton;		//button for creating a new document
	private Component usrSwitch;	//button for switching the current user
	private JTextField searchField;	//text field for searching for a certain document
	private JList itemList;			//list of items, each representing a document
	private JPopupMenu editPopUp;	//popup with option to open RTF version of a document
		
	//Previsualisation-related variables
	private LinkedHashMap<String, Icon> docPrevi;	//correlation between document and previz
	private CanvasDialogue diagPrevi;		        //container dialogue for previz image
	private JLabel previLabel,                      //container label used for previz image
				   previNameLabel;	                //text label with the document's full name
				
	//Integers
	private int iMaxChar,		//maximum number of characters allowed for text field
	            iClicked,		//index of list item selected with a mouse click 
	            iLastIndex,		//index of last list item the mouse was hovering over
	            iCurrentIndex,	//index of current list item the mouse is hovering over
	            iIdUsr;			//current user id
		
	//Otras variables... 
	private String sProfilePath, sPictPath;     	
	private LinkedHashMap<Object, Icon> itemsEmpty;	//items to display when text field is empty
	private LinkedHashMap<String, String> name2Id;  //correlation between document name and ID.
	private Timer cronoPrevi,						//timer for previz response	
	              cronoText;						//timer for text field response
	private Thread previThread;						//previz managing thread
	private static final int iIconDim = (int) (Global.ICON_SIZE * 1.75);
	
	/**
	 * Customised constructor. Sets up the dialogue with all the buttons, the item list,
	 * the previsualisation images and the text field.
	 * @param title Dialogue's title.
	 * @param iSpacing Spacing between list cells.
	 * @param iMaxChar Maximum number of characters for the text field
	 * @param editor Object representing the main editor window
	 * @param login Object representing the login dialogue
	 */
	public SelectDoc (String title, final int iSpacing, int iMaxChar, final Editor editor, final Login login) {
		
		//Empty dialogue with grey central and bottom backgrounds, non-resizable and "X" closes just the dialogue
		super (title, editor.getWindow(), null, Global.DEFAULT_BORDER, null, false, JFrame.DISPOSE_ON_CLOSE, 
			   Global.FAVICON_ED, false, false, true);
		
		String sIconFile,		//filename of image to be button icon 
		       sIconPath, 		//path to icon's image file 
		       sIdDoc;			//current document's ID
		ImageArray iconImage; 	//buffered image to be converted to icon 
		LinkedHashMap<String, String> userDocs;    //docName-to-docID list for all user's docs
		
		JMenuItem editRtfOp;	//popup menu item for the launch of the RTF file 
		JPanel centrePanel;		//dialogue's centre panel (where the item list goes)			   	
		JScrollPane listScroll; //scroll panel for list
		JLabel avatar;			//avatar image 
		
		//shows busy mouse pointer
		editor.busy (true);		
		
		//saves the current document
		editor.saveDoc();
		
		//initialises fields
		iIdUsr = editor.getUsrId();
		sProfilePath = Global.USERS_FOLDER +Global.SEPARADOR_SYS+ iIdUsr +Global.SEPARADOR_SYS;;
		sPictPath = Global.PICT_FOLDER +Global.SEPARADOR_SYS;
		this.editor = editor;	
		this.login = login;
		iClicked = -1;
		iLastIndex = -1;
		
		//initialises list-related variables
		itemsEmpty = new LinkedHashMap<Object, Icon>();
		name2Id = new LinkedHashMap<String, String>();
		userDocs = new LinkedHashMap<String, String>();
		
		//initialises previz managing thread
	    previThread = new Thread (this);	
		
		//initialises previz image label
		diagPrevi = new CanvasDialogue (null, getWindow(), null, null, Color.WHITE, false, JDialog.DO_NOTHING_ON_CLOSE, null, false); //totalmente desnuda
		diagPrevi.getWindow().setPreferredSize (new Dimension ((int) (Global.SCREEN_DIM.height / 3.5), 
				                                                (int) (Global.SCREEN_DIM.height / 2.625)));
		diagPrevi.getCanvas().setBorder (null); //Sin márgenes
		docPrevi = new LinkedHashMap<String, Icon>();
		previLabel = new JLabel();
		previLabel.setBorder (BorderFactory.createLineBorder (Global.DEFAULT_HIGHLIGHT, 4));
		previLabel.setHorizontalAlignment (SwingConstants.LEFT);
		previLabel.setVerticalAlignment (SwingConstants.TOP);
		previLabel.addMouseMotionListener (this);
		
		//initialises previz name label
		previNameLabel = new JLabel (" ");
		previNameLabel.setOpaque (true);
		previNameLabel.setBackground (Global.DEFAULT_HIGHLIGHT);
		JPanel previGlass = (JPanel) diagPrevi.getWindow().getGlassPane();
		previGlass.setLayout (new BorderLayout());
		previGlass.add (previNameLabel, BorderLayout.SOUTH);
		previGlass.setVisible (true);
		diagPrevi.populate (previLabel);
		previNameLabel.setPreferredSize (new Dimension (diagPrevi.getWindow().getWidth(), 
				                                        previNameLabel.getHeight() + 1));
				
		//sets up the "new document" button
		sIconPath = Global.IMG_FOLDER + Global.SEPARADOR_SYS;
		newButton = CanvasDialogue.toSwing (Global.ICON_SIZE, Global.ICON_SIZE, sIconPath, Global.NEWDOC, 
                                      Global.PICT_EXT, Global.LANG.getString ("nuevoBotonNuevo"), 
                                      this);		
		
		//sets up the "user switch" button if the login dialogue is linked to this dialogue
		if (login != null) 
		   usrSwitch = CanvasDialogue.toSwing (Global.ICON_SIZE, Global.ICON_SIZE, sIconPath, Global.USER_SWITCH, 
                                         Global.PICT_EXT, Global.LANG.getString ("nuevoBotonCambioUsr"), 
                                         this);		   
		else {
		     usrSwitch = new JLabel();
		     changeExitMode (JDialog.DO_NOTHING_ON_CLOSE);
		}
		
		//sets up the avatar image
		iconImage = new ImageArray (Global.USER_AVATAR, sProfilePath, Global.PICT_EXT);
		iconImage.resize (Global.ICON_SIZE, Global.ICON_SIZE);
		avatar = new JLabel (iconImage.toIcon());
		avatar.setToolTipText (editor.getUsrName());
		
		//sets up the text field
		searchField = new JTextField();
		searchField.addKeyListener (this);
		searchField.addMouseListener (this);
		this.iMaxChar = iMaxChar;
		
		//sets up the text field's timer for a response on key press
		cronoText = new Timer (Global.FIELD_RESPONSE, this);
		cronoText.setRepeats (false);
		
		//sets up the previz's timer for display on mouse hover
		cronoPrevi = new Timer (Global.PREVI_RESPONSE, this);
		cronoPrevi.setRepeats (false);
			
		//sets the list of default items and all related variables. 
		userDocs = Data.buscaItemsLista (iIdUsr);
		for (String sNombreDoc: userDocs.keySet()) {
			sIconFile = userDocs.get (sNombreDoc);
			sIdDoc = sNombreDoc.substring (0, sNombreDoc.lastIndexOf(Global.TOKEN_SEPARATOR));
			sNombreDoc = sNombreDoc.substring (sNombreDoc.lastIndexOf(Global.TOKEN_SEPARATOR) + 1);
						
			//prepares the previz image for the document in question
			iconImage = new ImageArray (sIdDoc, sProfilePath, Global.PICT_EXT);
			docPrevi.put (sNombreDoc, iconImage.toIcon());
			
			//if pictograms on for the user => display pictogram icon. If not, use previz as icon.
			if (editor.isPictosEnabled()) 
			   iconImage = new ImageArray (sIconFile, sPictPath, Global.PICT_EXT);
			else iconImage = new ImageArray (sIdDoc, sProfilePath, Global.PICT_EXT);
			iconImage.resize (iIconDim, iIconDim);
			itemsEmpty.put (sNombreDoc, iconImage.toIcon());
			name2Id.put (sNombreDoc, sIdDoc);
		}
				
		//Sets up each list cell as a pictogram with the document name below. 
		itemList = new JList (itemsEmpty.keySet().toArray());
		itemList.setCellRenderer (new DefaultListCellRenderer() {
            @Override
            public Component getListCellRendererComponent (JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
                JLabel label = (JLabel) super.getListCellRendererComponent (list, value, index, isSelected, cellHasFocus);
                Icon icon = itemsEmpty.get (value);
        	    label.setIcon (icon);
        	    label.setHorizontalAlignment (JLabel.CENTER);
        	    label.setHorizontalTextPosition (JLabel.CENTER);
        	    label.setVerticalTextPosition (JLabel.BOTTOM);
        	    label.setPreferredSize (new Dimension ((int) (Global.SCREEN_DIM.height / 6.3), 
										               (int) (Global.SCREEN_DIM.height / 8.75)));
        	    label.setBorder (BorderFactory.createEmptyBorder (iSpacing, iSpacing, iSpacing, iSpacing));
                return label;
            }
        });
		
		//Sets the Swing properties for the list
		itemList.setLayoutOrientation (JList.HORIZONTAL_WRAP);
		itemList.setVisibleRowCount (-1);
		itemList.addMouseMotionListener (this);
		itemList.addMouseListener (this);
		itemList.setCursor (Cursor.getPredefinedCursor (Cursor.HAND_CURSOR));
		listScroll = new JScrollPane (itemList);
		listScroll.setPreferredSize (new Dimension (12 * Global.ICON_SIZE, 5 * Global.ICON_SIZE));
				
		//Adds the buttons and text field to the top and bottom panels
		Component[] topLeft = new Component[1];
		topLeft[0] = newButton;
		Component[] topRight = new Component[2];
		topRight[0] = usrSwitch;
		topRight[1] = avatar;
		Component[] bottomRight = new Component[2];
		bottomRight[0] = new JLabel (Global.LANG.getString ("nuevoCampoTexto"));
		bottomRight[1] = searchField;
		
		//Adds the list to the centre panel
		centrePanel = CanvasDialogue.titledFrame (Global.LANG.getString ("nuevoListaTexto"), listScroll, null, true);
		centrePanel.addMouseMotionListener (this);
		
		//Adds all panels to the dialogue
		super.populate (topLeft, topRight, true, centrePanel, bottomRight);
		super.getWindow().addWindowListener (this);
		
		//Set up the RTF context menu
		if (Desktop.isDesktopSupported()) { 
		   Desktop desktop = Desktop.getDesktop();
		   if (desktop.isSupported (Desktop.Action.EDIT)) {
		      editPopUp = new JPopupMenu ();
		      editRtfOp = new JMenuItem (Global.LANG.getString ("nuevoRTFOp"));
		      editRtfOp.addActionListener (new ActionListener() {
		  	      public void actionPerformed (ActionEvent e) {
		  		      busy (true);
		  		      editor.launchRTF (name2Id.get(getSelected()), getSelected());
		  		      busy (false);
		  	      }
              });
		      editPopUp.add (editRtfOp);
		   }
		}
		
		//Resets mouse cursor
		this.editor.busy (false);		
	}
	
	/**
	 * Previz control thread. Displays the previz window and automatically closes it after 
	 * a fixed time interval.	 
	 */
	public void run() {
		diagPrevi.setVisible(true);
		try {
			Thread.sleep (Global.PREVI_INTERVAL);
		} catch (InterruptedException e) {
				diagPrevi.getWindow().setVisible (false);
		} 
		diagPrevi.getWindow().setVisible (false);
     }
	
	/**
	 * Returns the document name corresponding to the list item that has been clicked on.
	 * @return Document name. If there has not been a mouse click yet, an empty string is returned.
	 */
	public String getSelected () {
		if (iClicked == -1)
		   return "";
		else return (String) itemList.getModel().getElementAt (iClicked);
	}
	
	/**
	 * Manages all button events
	 */
	public void actionPerformed (ActionEvent e) {
		//Create a new document
		if (e.getSource() == newButton) {
		   cronoPrevi.stop();	
		   diagPrevi.close();
		   getWindow().setVisible (false);
		   if (editor.isPictosEnabled()) {
			  			  
			  //Builds and shows the document name with pictogram dialogue
			  NameDocPicto nombraDocPicto = new NameDocPicto (String.valueOf(iIdUsr));
	   		  boolean aceptado = nombraDocPicto.muestraCampoConfirmar (getWindow());
			  				  				  
			  //Acts upon confirming or cancelling document naming.
		      if (aceptado) {
		    	 if (editor.isDocOpen())
				    editor.closeDoc (false);
				 editor.newDoc (nombraDocPicto.getFields()[0], nombraDocPicto.getPictoId (Global.DEFAULT_PICT));
				 close();						 
		      } else setVisible(true);
		   
		   } else {
			      //Builds and shows the document name dialogue
			      NameDoc nombraDoc = new NameDoc (String.valueOf(iIdUsr));
			      boolean aceptado = nombraDoc.muestraCampoConfirmar (editor.getWindow());
			      
			      //Acts upon confirming or cancelling document naming
			      if (aceptado) {
			    	 if (editor.isDocOpen())
					    editor.closeDoc (false);
					 editor.newDoc (nombraDoc.getFields()[0], Global.DEFAULT_PICT);
					 close();
			      } else setVisible (true);	      			      
		   }
		   
		//Switching user
		} else if (e.getSource() == usrSwitch) {
			      cronoPrevi.stop();
			      diagPrevi.close();
			      close();
			      login.display();
			      
	    //Timer-triggered event: show the previz
		} else if (e.getSource() == cronoPrevi) {
			      if (iCurrentIndex == iLastIndex) 
			         previThread.start();			     
		}
	}
	
	/**
	 * Manages mouse moves over the item list area. When the mouse pointer is over a list item,
	 * the previz image of the document the item represents is displayed. 
	 */
	public void mouseMoved (MouseEvent e) {
		JDialog previWindow;  	    
		String sDocName;			//name of document represented by mouse-hovered list item
		Rectangle cellDim;			//cell area of one list item
		boolean bInsideCell;		//is the mouse pointer hovering over the list item's cell?
		
		previWindow = diagPrevi.getWindow();
		
		//Is the mouse pointer inside the list?
		if (e.getSource() == itemList) {	
		   iCurrentIndex = itemList.locationToIndex (new Point (e.getX(), e.getY()));
		   itemList.setSelectedIndex (iCurrentIndex);
		   
		   //Has it moved to another item? (Doesn't mean it's hovering over that item)
		   if (iCurrentIndex != iLastIndex) {  
			  
			  //Closes any popup menu still open
   	          if ((editPopUp != null) && editPopUp.isVisible())
   			     editPopUp.setVisible (false);	  
			   
			  //Stop any previous preview process 
			  cronoPrevi.restart();
			  previThread.interrupt();
  	          previThread = new Thread (this);
  	          
  	          //Find out the name of the document whose representative item has been highlighted
  	          iLastIndex = iCurrentIndex;
	          sDocName = (String) itemList.getModel().getElementAt (iCurrentIndex);
	          
	          //Find out the item's cell dimensions and find out if mouse indeed hovering over it.
	          cellDim = itemList.getCellBounds(iCurrentIndex, iCurrentIndex);
	          bInsideCell = cellDim.contains (e.getX(), e.getY());
	          
	          //Is there a previz available for the item mouse is hovering over?
	          if ((docPrevi.get(sDocName) != null) && (bInsideCell)){   
	    	     previLabel.setIcon (docPrevi.get(sDocName));
	    	     previNameLabel.setText ("<html><font color=\"#ffffff\">&nbsp;" +sDocName+ "</font></html>");
	    	    	    	     
	    	     //Positions the previz window relative to the list item's cell
	    	     previWindow.setLocation (getWindow().getX() + itemList.getX() + (int) cellDim.getMaxX(), 
	    	    		                 getWindow().getY() + itemList.getY() + (int) cellDim.getMaxY());
	          
	    	  //Mouse moves to a region of the list without items (happens when a row is partially
	    	  //filled with items) => close anything still open   
	          } else if (!bInsideCell) {
	        	        
	        	        //Although a list item has been highlighted, it's not below the mouse pointer
	        	        iLastIndex = -1;
	        	  
	        	        //Closes any preview window still open 
	        	        if (previWindow.isDisplayable()) {
	        	           previThread.interrupt();
	        	           previThread = new Thread (this);		                   
	        	        }
	        	        
	        	        //Closes any popup menu still open
	        	        if ((editPopUp != null) && editPopUp.isVisible())
	        			   editPopUp.setVisible (false);	        				        	        
	          }
		   } 	
		   
		//Mouse moves into the preview window (still open) => close preview 
	    } else if (previWindow.isDisplayable() || 
	    		  (previWindow.contains (new Point (e.getXOnScreen(), e.getYOnScreen())))) {  
	    	      previWindow.setVisible (false);
                  iLastIndex = -1;
	    }		
	}
	
	/**
	 * Acts upon clicking on a list item. Closes any currently displayed previz
	 * The selected document is opened and the previous one closed. 
	 */
	public void mouseClicked(MouseEvent mouse) {
		
		//Finds out if mouse is hovering exactly over one list item.		
		Rectangle cellDim = itemList.getCellBounds(iCurrentIndex, iCurrentIndex);
        if (cellDim.contains (mouse.getX(), mouse.getY())) {
           
           //Find out the name of the document represented by the list item being clicked
		   iClicked = itemList.locationToIndex (new Point (mouse.getX(), mouse.getY()));
		   		
   		   //Left-clicked on the item?
		   if (SwingUtilities.isLeftMouseButton(mouse)) {
		      editor.busy (true);
		      cronoPrevi.stop();
		   		   
		      //Closes all dialogues
		      diagPrevi.close();
		      super.close();
		      if (editor.isDocOpen())
	             editor.closeDoc (false);
		   
		      //Opens the document in question
		      try {
		          editor.openDoc (name2Id.get(getSelected()), getSelected());		       
		      } catch (Exception e) {
		              getWindow().getGlassPane().setCursor(Cursor.getPredefinedCursor (Cursor.DEFAULT_CURSOR));
	                  System.out.println ("Error when opening document: " +e);
	                  e.printStackTrace();
		      } finally {
			            editor.busy (false);
		      }
		 
		   //Right-clicked on item?   
	       } else if (SwingUtilities.isRightMouseButton(mouse) && (editPopUp != null)) {
	    	             	         
	    	         //Prevents the preview window from showing up and closes it if already present
	    	         previThread.interrupt();
	    	         cronoPrevi.stop();
	                     	      
	    	         //Displays the context menu
	    	         editPopUp.show ((JList) mouse.getSource(), mouse.getX(), mouse.getY());
	       }
        }
	}

	/**
	 * Manages the search of documents using the text field.
	 */
	public void keyReleased (KeyEvent e) {
    	LinkedHashMap<Object, Icon> items;		//list items matching search results	
    	LinkedHashMap<String, String> searched; //docName-to-docID list for searched docs
    	ImageArray iconImage;					//would-be icon image			
		String sIconFile;						//filename of list item's icon
		String sText;		                    //text from the field
    	    	   
    	//Nothing in the text field => show default item list
    	if (searchField.getText().isEmpty()) {
           itemList.setListData (itemsEmpty.keySet().toArray());
                   
        //Has the field's timer gone off => proceed to search  
        } else if (!cronoText.isRunning()) {
        	      searched = new LinkedHashMap<String, String>();
        	      
        	      //Shows busy mouse pointer
        	      busy (true);		
        	      
                  //Retrieves text from field and formats it
                  sText = searchField.getText();
                  if (sText.length() > iMaxChar) 
                	 sText = sText.substring (0, iMaxChar);
                  
                  //Performs the search and, if it yields results, changes the list 
                  searched = Data.buscaItemsLista (iIdUsr, sText.trim());
                  if ((searched != null) && !searched.isEmpty()) { 
                	 items = new LinkedHashMap<Object, Icon>();
                	 for (String sNombreDoc: searched.keySet()) {
                		 sIconFile = searched.get (sNombreDoc);
                		 
                		 //Prepares the previz image.
                		 iconImage = new ImageArray (name2Id.get (sNombreDoc), sProfilePath, Global.PICT_EXT);
             			 docPrevi.put (sNombreDoc, iconImage.toIcon());
             			
             			 //If pictograms on for the user => display pictogram icon. If not, use previz as icon.
              			 if (editor.isPictosEnabled()) 
              			    iconImage = new ImageArray (sIconFile, sPictPath, Global.PICT_EXT);
              			 else iconImage = new ImageArray (name2Id.get (sNombreDoc), sProfilePath, Global.PICT_EXT);              				    
              			 iconImage.resize (iIconDim, iIconDim);
              			 items.put (sNombreDoc, iconImage.toIcon());
              		}
                	 
                	//Displays searched documents 
                	itemList.setListData (items.keySet().toArray());
                                       
                  //If search invalid, fall back on default list
    	          } else itemList.setListData (itemsEmpty.keySet().toArray());
    	          cronoText.restart();
    	          
    	          //Resets mouse pointer
    	          busy (false);		
        }	 
    }
	
	/**
	 * Closes the previz window and the RTF popup menu if any is still open when the mouse 
	 * pointer moves to the text field.
	 */
	public void mouseEntered (MouseEvent e) {
		if (e.getSource() == searchField) {
		   if ((editPopUp != null) && editPopUp.isVisible()) 
			  editPopUp.setVisible (false);		   
   	       if (diagPrevi.getWindow().isVisible()) 
		      diagPrevi.getWindow().setVisible (false);	   
		}		
	}
	
	/**
	 * Closes any previz window still open when this dialogue is closed.
	 */
	public void windowClosing (WindowEvent e) {
		cronoPrevi.stop();
		if (diagPrevi.getWindow().isDisplayable())
		   diagPrevi.close();
    }
		    
	//Compulsory interface methods
	public void mouseExited (MouseEvent e) {}
	public void mouseDragged (MouseEvent e) {}
	public void mousePressed (MouseEvent e) {}
    public void mouseReleased (MouseEvent e) {}
    public void keyTyped (KeyEvent e) {}
    public void keyPressed (KeyEvent e) {}
    public void windowActivated (WindowEvent e) {}
	public void windowClosed (WindowEvent e) {}
	public void windowDeactivated (WindowEvent e) {}
	public void windowDeiconified (WindowEvent e) {}
	public void windowIconified (WindowEvent e) {}
	public void windowOpened (WindowEvent e) {}
	                                                                                   
}
