package gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.GraphicsEnvironment;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.Window;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;

import javax.imageio.ImageIO;
import javax.swing.AbstractButton;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.border.Border;
import javax.swing.border.TitledBorder;

import net.java.balloontip.BalloonTip;
import net.java.balloontip.styles.RoundedBalloonStyle;
import net.java.balloontip.utils.ToolTipUtils;
import utils.Global;

/**
 * Builds a panel (called canvas) split into three areas: top, centre and bottom. The top and bottom 
 * areas are both in turn halved into left and right sides, the latter being displayable in vertical 
 * (below the left side) or not displayed at all. Every element placed in the centre area will stretch
 * till the area is filled. Methods are also provided for basic operations common for both a
 * dialogues and windows proper.   
 * @author hqcasanova
 * 
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this file,
 * You can obtain one at http://mozilla.org/MPL/2.0/.
 */
public class Canvas {
	private JPanel canvas;						//main panel
    private JPanel topPanel, bottomPanel;		//top/bottom areas
    private Component centre;					//centre area
    private Window window;						//window (window proper or dialogue) the panel belongs to
        
    //Border around self-contained entities
    private static final Border spacing = BorderFactory.createEmptyBorder (5, 5, 5, 5);
    private static final int arcTooltip = 6;		//roundness of tooltip panel
    private static final int inTooltip = 500;		//time before tooltip is displayed
    private static final int outTooltip = 3000;		//time the tooltip keeps on display
    
    //Inner margins for titled button panel
    private static final int iBottomMargin = Math.round (Global.ICON_SIZE / 3.5f);
    private static final int iSideMargin = Math.round (Global.ICON_SIZE / 4.2f);
    private static final int iTopMargin = Math.round (Global.ICON_SIZE / 1.46f);

    /**
     * Customised constructor. Builds a new canvas with the specified background colours.
     * @param topPanelColour Background colour for top area.
     * @param bottomPanelColour Background colour for bottom area.
     * @param centrePanelColour Background colour for centre area. If colour value is null, it leaves the
     * default colour.
     * @param bRoundedTop True if top panel are to have rounded corners.
     * @param bRoundedBottom True if bottom panel are to have rounded corners.
     */
    public Canvas (Color topPanelColour, Color bottomPanelColour, Color centrePanelColour, boolean bRoundedTop, boolean bRoundedBottom) {
    	
    	//Sets up a new canvas with default spacing
    	canvas = new JPanel (new BorderLayout(0, 5)); 
    	canvas.setBorder (BorderFactory.createEmptyBorder (10, 10, 5, 10));
    
		//Sets up a new top panel with default spacing
    	if (bRoundedTop)
		   topPanel = new RoundedPanel (false, null);
    	else topPanel = new JPanel();
		if (topPanelColour != null) {  //Colour top panel?
		   topPanel.setBackground (topPanelColour);
		   topPanel.setBorder (spacing);
		}
		canvas.add (topPanel, BorderLayout.NORTH);
		
        //Sets up a new bottom panel with default spacing
		if (bRoundedBottom)
		   bottomPanel = new RoundedPanel (false, null);
		else bottomPanel = new JPanel();
		if (bottomPanelColour != null) {  //Colour bottom panel?
		   bottomPanel.setBackground (bottomPanelColour);
		   bottomPanel.setBorder (spacing);
		}
		canvas.add (bottomPanel, BorderLayout.SOUTH);
        
		//Changes the centre panel colour if this is provided.
        if (centrePanelColour != null)
 		   canvas.setBackground (centrePanelColour);
    }
    
    public Canvas (Color topPanelColour, Color bottomPanelColour, Color centrePanelColour) {
    	this (topPanelColour, bottomPanelColour, centrePanelColour, false, false);
    }
    
    /**
     * Sets the root container - the window
     * @param window
     */
    protected void setWindow (Window window) {
    	this.window = window;
    }
    
    /**
	 * Returns the main panel of the window for possible manipulation
	 * @return Main panel
	 */
    public JPanel getCanvas () {
    	return canvas;
    }
       
    /**
	 * Returns the top area of the window for further manipulation.
	 * @return Top panel
	 */
	public JPanel getTopPanel() {
		return topPanel;
	}
	
	/**
	 * Allows replacing the top area with a new panel.  
	 * @param newTop New top panel 
	 */
	protected void changeTopPanel (JPanel newTop) {
		canvas.remove (topPanel);
		topPanel = newTop;
		canvas.add (topPanel, BorderLayout.NORTH);
	}
	
	/**
	 * Removes the top area of the window. Particularly useful when reusing the canvas for windows
	 * without top toolbar. 
	 */
	public void removeTopPanel () {
		canvas.remove (topPanel);
	}
	
	 /**
	 * Returns the bottom area for further manipulation.
	 * @return Bottom panel
	 */
	protected JPanel getBottomPanel() {
		return bottomPanel;
	}
    
    /**
	 * Allows replacing the bottom area with a new panel.  
	 * @param newBottom New bottom panel 
	 */
	protected void changeBottomPanel (JPanel newBottom) {
		canvas.remove (bottomPanel);
		bottomPanel = newBottom;
		canvas.add (bottomPanel, BorderLayout.SOUTH);
	}
	
	/**
	 * Removes the bottom area of the window. Particularly useful when reusing the canvas for windows
	 * without confirmation buttons. 
	 */
	public void removeBottomPanel () {
		canvas.remove (bottomPanel);
	}
	
	/**
	 * Retrieves the centre area for further manipulation
	 * @return Container component of elements in centre area
	 */
	protected Component getCentre() {
		return centre;
	}
	
	/**
	 * Allows replacing the centre area with a new container
	 * @param newCentre New centre component.
	 */
	protected void changeCentre (Component newCentre) {
		canvas.remove (centre);
		canvas.add (newCentre, BorderLayout.CENTER);
		window.pack();
	}
	
	/**
	 * Adds components to the appropriate area of the canvas.
	 * @param topLeftComps Components to be placed in top left section. 
	 * @param topRightComps Components to be placed in top right section. Allowed to be null.
	 * @param horizontal Top components to be laid out vertically?
	 * @param centre Components to be placed in the centre area.
	 * @param bottomLeftComps Components for the bottom left section. If null, components in
	 * the bottom right section will be guaranteed to right align. 
	 * @param bottomRightComps Components for the bottom right section. 
	 */
	public void populate (Component[] topLeftComps, Component[] topRightComps, boolean horizontal, 
			            Component centre, Component[] bottomLeftComps, Component[] bottomRightComps) {
		JPanel topLeftPanel, topRightPanel;  		
		JPanel bottomLeftPanel, bottomRightPanel;  	
		int i;
		
		this.centre = centre;
		try {			
			
			//Sets up top panels 
			topLeftPanel = new JPanel ();
			topLeftPanel.setOpaque (false);   
			topRightPanel = new JPanel ();
			topRightPanel.setOpaque (false);
			
			//Two horizontal cells with opposite justification
			if (horizontal) { 
					   
			   topPanel.setLayout (new GridLayout (0, 2));
			   topLeftPanel.setLayout (new FlowLayout (FlowLayout.LEADING));
			   topRightPanel.setLayout (new FlowLayout (FlowLayout.TRAILING));
			
			//Two vertical cells with left justification   
			} else { 
				   topPanel.setLayout (new GridLayout (2, 0));
				   topLeftPanel.setLayout (new BorderLayout());
				   topRightPanel.setLayout (new BorderLayout());
			}
			
			//Adds components to the top area
			for (i = 0; i < topLeftComps.length; i++)
	    	    topLeftPanel.add (topLeftComps[i]);
	        topPanel.add (topLeftPanel);
	        if (topRightComps != null) { 
	           for (i = 0; i < topRightComps.length; i++)
	    	       topRightPanel.add (topRightComps[i]);
	           topPanel.add (topRightPanel);
	        }
	            
	        //Adds components to the centre area
	        canvas.add (centre, BorderLayout.CENTER);
	    
	        //Sets up bottom panels
	        bottomRightPanel = new JPanel ();
	        bottomRightPanel.setOpaque (false);
	        
	        //Components for top left panel? => set top panels with opposite justification
	        if (bottomLeftComps != null) { 
	           bottomPanel.setLayout (new GridLayout (0, 2));
	           bottomRightPanel.setLayout (new FlowLayout (FlowLayout.TRAILING));
	           bottomLeftPanel = new JPanel ();
	           bottomLeftPanel.setOpaque (false);
	           bottomLeftPanel.setLayout (new FlowLayout (FlowLayout.LEADING));
	           for (i = 0; i < bottomLeftComps.length; i++)
	    	       bottomLeftPanel.add (bottomLeftComps[i]);
	           bottomPanel.add (bottomLeftPanel);
	        
	        //Set bottom panel justification to right
	        } else {  
	        	   bottomPanel.setLayout (new BorderLayout());
	        	   bottomRightPanel.setLayout (new BorderLayout());
	        }	     
	        
	        //Add components to bottom right section
	        bottomRightPanel.add (bottomRightComps[0], BorderLayout.WEST);
	        bottomRightPanel.add (bottomRightComps[1], BorderLayout.CENTER);
	        bottomPanel.add (bottomRightPanel);	
	        
	        //Compress the window removing extra space
	        window.pack();
	       
		} catch (Exception e) {
			    System.out.println ("Error al crear ventana: " +e);
			    e.printStackTrace();
		}
	}
	
	/**
	 * Overloaded method for cases without top and bottom areas
	 * @param centre Components for centre area	 
	 */
	public void populate (Component centre) {
		this.centre = centre;
		removeTopPanel();
		removeBottomPanel();
		canvas.add (centre, BorderLayout.CENTER);
		window.pack();
    }
	
	/**
	 * Overloaded method for cases without partitioned top and bottom areas
	 */
	public void populate (Component top, Component centre, Component bottom) {
		canvas.add (top, BorderLayout.NORTH);
		canvas.add (bottom, BorderLayout.SOUTH);
		populate (centre);
    }
	
	/**
	 * Overloaded method for cases without top-right section. As there is just one top section,
	 * the layout is assumed to be horizontal.
	 */
	public void populate (Component[] topComps, Component centre,
			            Component[] bottomLeftComps, Component[] bottomRightComps) {
		populate (topComps, null, true, centre, bottomLeftComps, bottomRightComps);
	}
	
	/**
	 * Overloaded method for cases without bottom-left section. 
	 */
	public void populate (Component[] topLeftComps, Component[] topRightComps, boolean horizontal, 
			            Component centre, Component[] bottomComps) {
		populate (topLeftComps, topRightComps, horizontal, centre, null, bottomComps);
    }
		
	/**
	 * Encloses in a titled border frame a given group of components. These are to be positioned in 
	 * the frame's centre and bottom sections. 
	 * @param sTitle Border frame title
	 * @param centre Component for centre section of the frame.
	 * @param bottom Component for bottom section of the frame.
	 * @return Returns a panel with everything set up. 
	 */
	protected static JPanel titledFrame (String sTitle, Component centre, Component bottom, boolean bRound) {
		JPanel panel;
		if (!bRound)
		   panel = new JPanel ();
		else panel = new RoundedPanel (false, null);
		panel.setLayout (new BorderLayout(0, 3));
				
		panel.add (new JLabel (sTitle), BorderLayout.NORTH);
		if (centre != null)
		   panel.add (centre, BorderLayout.CENTER);	
		if (bottom != null)
		   panel.add (bottom, BorderLayout.SOUTH);
		panel.setBorder (spacing);
		panel.setBackground (Global.DEFAULT_BORDER); 
		return panel;
	}
	
	protected static JPanel titledFrame (String sTitle, Component centre, Component bottom) {
		return titledFrame (sTitle, centre, bottom, false);
	}
	
	/**
	 * Surrounds a given panel made up of buttons with a titled border
	 * @param group Panel with buttons
	 * @param sTitle Title to be shown on border
	 * @param vertical True if the panel's layout axis is vertical. False, if not.
	 * @return Panel with titled border.
	 */
	public static JPanel titledButtonPanel (JPanel group, String sTitle, boolean vertical) {
		TitledBorder border; //border surrounding buttons
		
		//Removes the space after the last added button
		group.remove (group.getComponentCount() - 1);
		  
		//Sets the title of the group
		if (!sTitle.isEmpty()) {
		   border = new TitledBorder (sTitle) {
		    	            public Insets getBorderInsets (Component c) {
		    		            return new Insets (iTopMargin, iSideMargin, iBottomMargin, iSideMargin);
		    	            }
		            }; 
		   if (vertical || (group.getComponentCount() == 1))
			  border.setTitleJustification (TitledBorder.CENTER);
		   else border.setTitleJustification (TitledBorder.LEFT);
		   group.setBorder (border);		     
		}
		  
		//Position newly created group centred relative to other existing elements
	    group.setAlignmentX (Component.CENTER_ALIGNMENT);
	    group.setAlignmentY (Component.CENTER_ALIGNMENT);
	    
	    return group;
	}
	
	
		
	/**
	 * Removes all buttons present in the window. The close and minimised buttons are also removed if
	 * the look & feel in used is native to java. 
	 * @param root Root container (normally the window itself)
	 */
	protected static void removeButtons (Component root) {
		int x, y;
		
	    if (root instanceof AbstractButton)
	       root.getParent().remove (root);
	    if (root instanceof Container) {  
	       Component[] comps = ((Container)root).getComponents();  
	       for (x = 0, y = comps.length; x < y; x++) {  
	           removeButtons (comps[x]);  
	       }  
	    }  
	}
		
	/**
	 * Sets the favicon for the window
	 * @param sFavicon Relative path for the image to use as favicon. If null, no favicon is set. 
	 */
	protected void setIcon (String sFavicon) {
		try {
			if (sFavicon != null)
			   window.setIconImage (ImageIO.read (new File (sFavicon)));
		} catch (IOException e) {
			    System.out.println ("Error al fijar el favicon " +sFavicon+ ": " +e);
			    e.printStackTrace();
		}
	}
	
	/**
	 * Forces the packing, relative positioning and painting of the window.
	 * @param relative Container relative to which the window is positioned.
	 */
	public void display (Component relative) {
		window.setMinimumSize (new Dimension (window.getWidth(), window.getHeight()));
		window.setLocationRelativeTo (relative);
		window.setVisible (true);
	}
	
	/**
	 * Overloaded method with automatic centering of the window when displayed
	 */
	public void display () {
		display (null);
	}
	
	/**
	 * Refreshes the contents of the window. Useful when components have been removed or added.
	 */
	protected void refresh () {
		window.pack();
		window.repaint();
	}
	
	/**
	 * Makes the window visible
	 * @param visible True if visible. False if not.
	 */
	public void setVisible (boolean visible) {
		window.setVisible (visible);
	}
	
	/**
	 * Determines if the window is visible (has been rendered on the display).
	 * @return True if it has been rendered. False if not.
	 */
	public boolean isVisible () {
		return window.isVisible();
	}
	
	/**
	 * Closes the window.
	 */
	public void close() {
		window.dispose();
	}
	
	/**
	 * Modifies the font of all components in the window
	 * @param newFont New font for components.
	 */
	public void setFontAll (Font newFont) {
	    setFuenteRecursive (canvas.getComponents(), newFont);
	}

	/**
	 * Recursively changes the font of all children components of a given container 
	 * @param components Container's components
	 * @param font Font to use for all components.
	 */
	private static void setFuenteRecursive (Component[] components, Font font) {
	    for (Component c : components) {
	        c.setFont (font);
	        if (c instanceof Container)
	            setFuenteRecursive (((Container)c).getComponents(), font);
	    }
	}
	
	/**
	 * Adds a balloon-shaped tooltip to a given component
	 * @param component Component the tooltip is added to
	 * @param sTip Tooltip text
	 */
	public static void setTooltip (JComponent component, String sTip) {
		BalloonTip bocadillo = new BalloonTip (component, sTip, new RoundedBalloonStyle (arcTooltip, arcTooltip, Global.TOOLTIP_BACKGROUND, Global.DEFAULT_HIGHLIGHT), false);
		ToolTipUtils.balloonToToolTip (bocadillo, inTooltip, outTooltip);	
	}
	
	/**
	 * Sets up a JButton object with the properties specified.
	 * @param iIconWidth Button's icon width
	 * @param iIconHeight Button's icon height
	 * @param sIconPath Path to icon image file.
	 * @param sIconName Filename of icon image.
	 * @param sExtension Icon file extension
	 * @param sToolTip Tooltip text on hovering the button.
	 * @param listener Class managing events from button
	 * @return Swing JButton object
	 */
	public static JButton toSwing (int iIconWidth, int iIconHeight, String sIconPath, 
			                       String sIconName, String sExtension, 
			                       String sTooltip, ActionListener listener) {
		JButton button;
		ImageArray bufferedImage;
		
		//Sets button's icon
		bufferedImage = new ImageArray (sIconName, sIconPath, sExtension);
		bufferedImage.resize (iIconWidth, iIconHeight);
		button = new JButton (bufferedImage.toIcon());
				
		//Sets dynamic properties
		Canvas.setTooltip (button, sTooltip);
		button.addActionListener (listener);
		button.setActionCommand (sIconName);
		button.setCursor (Cursor.getPredefinedCursor (Cursor.HAND_CURSOR));
		return button;		
	}
	
	/**
	 * Sets up a JComboBox object representing a drop-down menu. It also labels the menu with
	 * some text on top of it.
	 * @param dropDownItems Items to be displayed in menu
	 * @param listener Class managing events from menu
	 * @param sActionCmd Action command issued when interacting with the Swing object
	 * @param sLabel Title text
	 * @return Swing JComboBox object
	 */
	public static JPanel toSwing (String[] dropDownItems, ActionListener listener, 
			                      String sActionCmd, String sLabel) {
		JComboBox dropDown;
		JLabel label;
		JPanel panel;
		
		dropDown = new JComboBox (dropDownItems);
	    dropDown.addActionListener (listener);
	    dropDown.setActionCommand (sActionCmd);
	    dropDown.setAlignmentX (Component.LEFT_ALIGNMENT);
	    panel = new JPanel ();
	    panel.setLayout (new BoxLayout (panel, BoxLayout.Y_AXIS));
	    label = new JLabel (sLabel);
	    label.setAlignmentX (Component.LEFT_ALIGNMENT);
	    panel.add (label);
	    panel.add (Box.createVerticalGlue());
	    panel.add (dropDown);
	    return panel;
	}
	
	/**
	 * Get correct work area bounds according to system (workaround for bug 4737732)
	 * http://bugs.sun.com/view_bug.do?bug_id=4737732 
	 */
	public static Rectangle getWorkArea() {
		if (GraphicsEnvironmentUtil.isX11GraphicsEnvironment) {
		   Rectangle workArea = GraphicsEnvironmentUtil.getX11RootNetWorkarea();
		   
		   //The worked out dimensions are not flawed (on Ubuntu, sometimes a very small value is given
		   //as the inset at the bottom of the desktop (where the task bar normally sits)
		   if ((Global.SCREEN_DIM.height - workArea.height) > 1) 
		      return workArea;
		   
		   //Dimensions are flawed => no other option but to assume a 50-pixel taskbar.
		   else {
			    workArea.height = Global.SCREEN_DIM.height - 50;
			    return workArea;
		   }
		} else return GraphicsEnvironment.getLocalGraphicsEnvironment().getMaximumWindowBounds();			
	}
}
